
#include "gft.h"

#include <cstring>

#define IMG_COLOR     1
#define IMG_GRAYSCALE 0

gft::CImage::CImage *ReadAnyCImage(char *file){
  gft::CImage::CImage *cimg;
  char command[512];
  int s;

  s = strlen(file);
  if(strcasecmp(&file[s-3], "ppm") == 0){
    cimg = gft::CImage::Read(file);
  }
  else{
    sprintf(command, "convert %s cimage_tmp.ppm", file);
    system(command);

    cimg = gft::CImage::Read((char *)"cimage_tmp.ppm");
    system("rm cimage_tmp.ppm");
  }
  return cimg;
}


gft::Image32::Image32 *ReadAnyImage(char *file){
  gft::Image32::Image32 *img;
  char command[512];
  int s;

  s = strlen(file);
  if(strcasecmp(&file[s-3], "pgm") == 0){
    img = gft::Image32::Read(file);
  }
  else{
    sprintf(command, "convert %s image_tmp.pgm", file);
    system(command);
    img = gft::Image32::Read((char *)"image_tmp.pgm");
    system("rm image_tmp.pgm");
  }
  return img;
}


int main(int argc, char **argv){
  gft::CImage::CImage *cimg=NULL, *spixels, *tmp;
  gft::Image32::Image32 *img=NULL, *label;
  char filename[512];
  int k, color = 0xFF0000, black=0x000000, type;
  double alpha;
  bool isMixed, isRoot;

  if(argc != 8){
    fprintf(stdout,"usage:\n");
    fprintf(stdout,"superpixels <type> <image> <k> <alpha> <sampling> <cost-function> <output>\n");
    fprintf(stdout,"\t type..... 0: grayscale\n");
    fprintf(stdout,"\t           1: color\n");
    fprintf(stdout,"\t k........ The desired number of superpixels\n");
    fprintf(stdout,"\t alpha.... The relative importance between color similarity and spatial proximity.\n");
    fprintf(stdout,"\t           It can be in the range [0.01, 0.50]\n");
    fprintf(stdout,"\t sampling. 0: grid\n");
    fprintf(stdout,"\t           1: mixed\n");
    fprintf(stdout,"\t cost..... 0: mean\n");
    fprintf(stdout,"\t           1: root\n");
    fprintf(stdout,"\t output... The target path of the output sans extension.\n");
    exit(0);
  }

  type = atoi(argv[1]);
  strcpy(filename, argv[2]);
  k = atoi(argv[3]);
  alpha = atof(argv[4]);
  isMixed = atoi(argv[5]);
  isRoot = atoi(argv[6]);

  if(type == IMG_GRAYSCALE){
    img  = ReadAnyImage(filename);
    cimg = gft::CImage::Clone(img);
  }
  else
    cimg = ReadAnyCImage(filename);


  //-------------------------------------------------
  
  float real_proc_time = 0.0;
  if(type == IMG_GRAYSCALE)  
    label = gft::Superpixels::ISF(img, k, alpha, //12.0,
				       5.0, 2.0, 10, isMixed, isRoot, &real_proc_time);
  else
    label = gft::Superpixels::ISF(cimg, k, alpha, //12.0,
				       5.0, 2.0, 10, isMixed, isRoot, &real_proc_time);

  tmp = gft::Highlight::CWideLabels(cimg, label, 2.0, &black, 0.0,
				    true, true);
  spixels = gft::Highlight::CWideLabels(tmp, label, 1.0, &color, 0.0,
					true, true);
  gft::CImage::Destroy(&tmp);

  printf("N = %d superpixels\n", gft::Superpixels::GetNumberOfSuperPixels(label));

  sprintf(filename, "%s.pgm", argv[7]);
  gft::Image32::Write(label, filename);
  sprintf(filename, "%s.ppm", argv[7]);
  gft::CImage::Write(spixels, filename);

  gft::Image32::Destroy(&label);
  gft::CImage::Destroy(&spixels);
  if(cimg != NULL) gft::CImage::Destroy(&cimg);
  if(img != NULL)  gft::Image32::Destroy(&img);
  return 0;
}



